//
//  SysctlTask.h
//  SecurityTutorial
//
//  Created by brian on Sat Mar 24 2001.
//  Copyright (c) 2001 Brian R. Hill. All rights reserved.
//

#import <Cocoa/Cocoa.h>
#import <Security/Authorization.h>

@interface SysctlTask : NSObject {
	AuthorizationRef authorizationRef; 
}
+ sharedInstance;

// Execute the sysctl tool with the given arguments using
// AuthorizationExecuteWithPrivileges.
// Attempt to authorize the user if necessary.
//
- (void)setValue:(char *)inValue;


// Find out if the user has the appropriate authorization rights.
// This really needs to be called each time you need to know whether the user
// is authorized, since the AuthorizationRef can be invalidated elsewhere, or
// may expire after a short period of time.
- (BOOL)isAuthenticated;


// if the user is not already authorized, ask them to authenticate.
// return YES if the user is (or becomes) authorized.
//
- (BOOL)authenticate;


// Invalidate the user's AuthorizationRef to dispose of any acquired rights.
// It's a good idea to do this before you quit your application, else they
// will stay authorized as of the current time (4K78).
//
- (void)deauthenticate;
@end

// SysctlTask sends these notifications are sent when the user  
// becomes authenticated or deauthenticates.
extern NSString* SysctlAuthenticatedNotification;
extern NSString* SysctlDeauthenticatedNotification;




